/** @file       log.h
    @brief
    @details
    @version    1.0.0
*/
#ifndef LOG_H_
#define LOG_H_

#ifdef __cplusplus
extern "C" {
#endif

#ifndef __KERNEL__
#include <string.h>
#include <time.h>
#else
#include <linux/string.h>
#endif

//#define TIMEZONE_EAST_8H        (8 * 3600)

#ifndef __KERNEL__
#define DUMP_PREFIX_NONE        (0)
#define DUMP_PREFIX_ADDRESS     (1)
#define DUMP_PREFIX_OFFSET      (2)
#endif

#define print(format, ...)   fprintf(stdout, format, __VA_ARGS__)


/** @enum _LOG_LEVEL_
 *  LOG等级枚举变量
 */
typedef enum {
    LOG_Fatal   = (1u << 0u),
    LOG_Error   = (1u << 1u),
    LOG_Warn    = (1u << 2u),
    LOG_Debug   = (1u << 3u),
    LOG_Info    = (1u << 4u),
    LOG_Test    = (1u << 5u),
    LOG_Call    = (1u << 6u),
    LOG_Devp    = (1u << 7u),
    LOG_Step    = (1u << 8u),
    LOG_Unknown = (1u << 9u),
    LOG_All     = (0x7FFFFFFF),
    LOG_Close   = 0x0,
} LOG_LEVEL;
/** @var typedef _LOG_LEVEL_ LOG_LEVEL
 *  @brief 错误值枚举类型
 */

#ifdef DISABLE_LOG
#define LOG_BUF(level, buf, len)
#define LOG_EX(level, format, args...)
#define LOG_EX2(level, format, args...)
#define LOG_TAG_EX(tag, level, format, args...)
#define DEBUG_CODE_LINE()
#define DEBUG_FUNCTION_BEGIN()
#define DEBUG_FUNCTION_END()
#else
#define LOG_BUF(level, buf, len, prefix) do { \
        const char* pFmtBuf = format_hex_buf(prefix, DUMP_PREFIX_NONE, 16, 1, buf, len, 1); \
        IHW_LOG(level, "[%s] - %s(%04d): %s", basename_v2(__FILE__), __FUNCTION__, __LINE__, pFmtBuf); \
        free((void*)pFmtBuf); \
    } while(0);
/*! \def LOG_EX
    \brief 系统日志调试宏标识
*/
#define LOG_EX(level, format, args...) (IHW_LOG(level, "[%s] - %s(%04d):" format , basename_v2(__FILE__), __FUNCTION__, __LINE__, ##args))

/*! \def LOG_TAG_EX
    \brief 系统日志调试宏标识
*/
#define LOG_TAG_EX(tag, level, format, args...) (IHW_LOG(level, "{%s} [%s] %s(%04d):" format , tag, basename_v2(__FILE__), __FUNCTION__, __LINE__, ##args))

#define LOG_EX2(level, format, args...) (IHW_LOG_UNTAG(level, format , ##args))

/*! @def APP_BUILD_INFO
    @brief 应用程序编译信息
*/
#define APP_BUILD_INFO(appname, appver) (IHW_LOG(LOG_Info, "%s Ver:%s (Build: %s %s  GCC Ver:%s) With %d(bits) OS\n", \
                                                 appname, appver, __DATE__, __TIME__, __VERSION__, sizeof(int*) * 8))


/*! @def DEBUG_CODE_LINE
    @brief 输出当前函数名，行号
*/
#define DEBUG_CODE_LINE()           (LOG_EX(LOG_Info, "\n"))

/*! @def DEBUG_FUNCTION_BEGIN
    @brief 函数入口标志
*/
#define DEBUG_FUNCTION_BEGIN()      (LOG_EX(LOG_Call, "+++++\n"))

/*! @def DEBUG_FUNCTION_END
    @brief 函数出口标志
*/
#define DEBUG_FUNCTION_END()        (LOG_EX(LOG_Call, "-----\n"))

/**
 * @brief           输出调试信息
 * @param level     调试信息开关
 * @param pMsg      调试信息内容
 */
void IHW_LOG(unsigned int level, const char *pMsg, ...);
void IHW_LOG_UNTAG(unsigned int level, const char *pMsg, ...);

void IHW_LogStrWithoutPrint(unsigned int level, char *pMsg);
void IHW_LogRawString(unsigned int level, char *pMsg);

/**
 * @brief           设置调试等级
 * @param level     调试等级
 * @param iEnable   1 打开调试等级， 0 关闭调试等级
 */
void IHW_EnableLogLevel(unsigned int level, int iEnable);

/**
 * @brief           初始化系统日志功能
 * @param pLogTag   系统日志标志
 * @param pPath     系统日志保存路径
 * @param bEnable   打开/关闭调试信息
 */
void IHW_InitLOG(const char *pLogTag, const char *pPath, int bEnable);

void IHW_RunLogService(void);

/**
 * @brief 判断文件、路径是否存在
 * @param pPath - 文件路径
 * @return int 存在返回 1, 否则返回 0;
 */
int IHW_IsFileExist(const char *pPath);

char *IHW_bin2hex(char *p, const unsigned char *cp, int count);

/* Return the last part of a pathname */
static inline const char *basename_v2(const char *path)
{
    const char *tail = strrchr(path, '/');
    return tail ? tail + 1 : path;
}

static inline int dirname_v2(const char *path, char *dir)
{
    const char *tail = strrchr(path, '/');

    if(tail) {
        memcpy(dir, path, tail - path);
        dir[tail - path] = 0;
    } else {
        strcpy(dir, "./");
    }

    return 0;
}
#endif

void IHW_WaitFinishLogout(void);
const char *LogLeveToString(unsigned int lv);

const char *format_hex_buf(const char *prefix_str, int prefix_type,
                           int rowsize, int groupsize,
                           const void *buf, int len, int ascii);
#ifndef __KERNEL__
void print_hex_dump_bytes(const char *prefix_str, int prefix_type,
                          const void *buf, int len);

#endif

#ifdef __cplusplus
}
#endif
#endif //LOG_H_

